#define _WIN32_DCOM

#include <windows.h>
#include "../Commun/UUID_Fabriques.h"
#include "../Commun/CompterTemps_h.h"
#include <string>
#include <iostream>
#include <ChargeurCom.h>
#include <ConversionChaines.h>

const char CHOIX_TICSPARUNITE = 'T',
           CHOIX_UNITESPARETAPE = 'U',
           CHOIX_QUITTER = 'Q';

// Affiche la configuration courante de l'horloge
void AfficherConfigCourante (IConfigTemps *);

// Affiche le menu de cet utilitaire
void AfficherMenu ();

// Retourne VRAI si le choix pass en paramtre est
// un choix valide, FAUX sinon
bool ChoixEstValide (const char);

// Lit et retourne un choix (non valid)
char LireChoix ();

// Lit et retourne un entier long strictement positif
long LireEntierPositif (const std::string & Invite = "");

// Lit et valide un choix, puis le retourne
char SaisirChoix ();

// Traite le choix reu en paramtre (valid) 
// l'aide du pointeur de IConfigTemps (valide)
void TraiterChoix (IConfigTemps *, const char);

// Fixe les paramtres de connexion au serveur
class OutilDeConnexion
{
   static const std::wstring LOOPBACK_IP, LOOPBACK_TEXTE;
   static const std::wstring CTX_LOCAL, CTX_DISTANT;
   static const COSERVERINFO COSERVERINFO_Defaut;
   static const COAUTHINFO COAUTHINFO_Defaut;
   std::wstring m_IpServeur;
   long m_CtxServeur;
   COAUTHINFO m_InfoAuth;
   COSERVERINFO m_InfoServeur;
   bool ContexteLocal () const { return GetContexteServeur () == CLSCTX_LOCAL_SERVER; }
   bool ContexteDistant () const { return GetContexteServeur () == CLSCTX_REMOTE_SERVER; }
public:
   long GetContexteServeur () const { return m_CtxServeur; }
   const std::wstring & GetDescrContexteServeur () const
   {
      return ContexteLocal ()? CTX_LOCAL : CTX_DISTANT;
   }
   std::wstring GetIp () const { return m_IpServeur; }
   COSERVERINFO *GetInfoServeur () { return &m_InfoServeur; }
   OutilDeConnexion (const std::wstring &Ip)
      : m_IpServeur (Ip.length ()? Ip : LOOPBACK_IP),
        m_CtxServeur (Ip.length ()? CLSCTX_REMOTE_SERVER : CLSCTX_LOCAL_SERVER),
        m_InfoAuth (COAUTHINFO_Defaut),
        m_InfoServeur (COSERVERINFO_Defaut)
   {
      m_InfoAuth.dwAuthnLevel = RPC_C_AUTHN_LEVEL_NONE;
      m_InfoAuth.dwAuthnSvc = RPC_C_AUTHN_NONE;
      m_InfoAuth.dwAuthzSvc = RPC_C_AUTHZ_NONE;
      m_InfoAuth.dwCapabilities = EOAC_NONE;
      m_InfoAuth.dwImpersonationLevel = RPC_C_IMP_LEVEL_IMPERSONATE;
      m_InfoServeur.pAuthInfo = &m_InfoAuth;
      m_InfoServeur.pwszName = const_cast<wchar_t*>(m_IpServeur.c_str ());
   }
};
const std::wstring
   OutilDeConnexion::LOOPBACK_IP = L"127.0.0.1",
   OutilDeConnexion::LOOPBACK_TEXTE = L"localhost";
const std::wstring
   OutilDeConnexion::CTX_LOCAL = L"CLSCTX_LOCAL_SERVER",
   OutilDeConnexion::CTX_DISTANT = L"CLSCTX_REMOTE_SERVER";
const COSERVERINFO
   OutilDeConnexion::COSERVERINFO_Defaut = { 0 };
const COAUTHINFO
   OutilDeConnexion::COAUTHINFO_Defaut = { 0 };


int main (int argc, char *argv[])
{
   ChargeurCOM ccom;

   // --- CONNEXION
   OutilDeConnexion odc (argc > 1 ? StringToWString (argv[1]) : L"");
   std::wcout << L"Tentative de connexion au serveur..." << std::endl
              << L"--> Adresse:  " << odc.GetIp () << std::endl
              << L"--> Contexte: " << odc.GetDescrContexteServeur () << std::endl;

   // --- SCURIT
   // Scurit minimum
   ::CoInitializeSecurity(0, -1, 0, 0, RPC_C_AUTHN_LEVEL_NONE,
                          RPC_C_IMP_LEVEL_IDENTIFY, 0, EOAC_NONE, 0);

   // --- CONNEXION ET INSTANCIATION
   // On ne demande qu'une seule interface: IConfigTemps
   HRESULT hr;
   MULTI_QI mqi [1] = { { &IID_IConfigTemps, 0, S_OK } };
   hr = ::CoCreateInstanceEx (CLSID_FabriqueServeurTempsQuiPasse, 0,
                              odc.GetContexteServeur (),
                              odc.GetInfoServeur (), 1, mqi);
   if (FAILED (hr))
   {
      std::cerr << "Erreur de fabrication du serveur. Code: " << hr
                << std::endl;
      return -1;
   }

   IConfigTemps *pConfigTemps = static_cast<IConfigTemps *>(mqi[0].pItf);

   // Utilisation de l'interface
   AfficherConfigCourante (pConfigTemps);
   char Choix = SaisirChoix ();
   while (Choix != CHOIX_QUITTER)
   {
      TraiterChoix (pConfigTemps, Choix);
      AfficherConfigCourante (pConfigTemps);
      Choix = SaisirChoix ();
   }
   pConfigTemps->Release ();
}


// Affiche la configuration courante de l'horloge
void AfficherConfigCourante (IConfigTemps *pConfigTemps)
{
   long TicsParUnite,
        UnitesParEtape;

   pConfigTemps->GetUniteDeTemps(&TicsParUnite);
   pConfigTemps->GetUnitesParEtape(&UnitesParEtape);
   std::cout << std::endl << std::endl << std::endl
             << "Horloge -- Configuration courante" << std::endl
             << "Nombre de tics par unite de temps: "
             << TicsParUnite
             << std::endl
             << "Nombre d'unites de temps par tape: "
             << UnitesParEtape
             << std::endl;
}

// Affiche le menu de cet utilitaire
void AfficherMenu ()
{
   std::cout << "---------------------" << std::endl
             << "[" << CHOIX_TICSPARUNITE << "] Modifier les tics par unite de temps?" << std::endl
             << "[" << CHOIX_UNITESPARETAPE << "] Modifier les unites de temps par tape?" << std::endl
             << "[" << CHOIX_QUITTER << "] Quitter?" << std::endl
             << "---------------------" << std::endl;
}

// Retourne VRAI si le choix pass en paramtre est
// un choix valide, FAUX sinon
bool ChoixEstValide (const char Choix)
{
   return Choix == CHOIX_TICSPARUNITE ||
          Choix == CHOIX_UNITESPARETAPE ||
          Choix == CHOIX_QUITTER;
}

// Lit et retourne un choix (non valid)
char LireChoix ()
{
   char Choix;
   std::cin >> Choix;
   return toupper (Choix);
}

// Lit et retourne un entier long strictement positif
long LireEntierPositif (const std::string & Invite)
{
   const std::string INVITE =
      Invite.length ()? Invite : "Entier strictement positif? ";
   std::cout << INVITE;
   long ValeurLue;
   std::cin >> ValeurLue;
   while (ValeurLue <= 0)
   {
      std::cout << "La valeur doit tre strictement positiive..."
                << std::endl << INVITE;
      std::cin >> ValeurLue;
   }
   return ValeurLue;
}

// Lit et valide un choix, puis le retourne
char SaisirChoix ()
{
   char Choix;
   do
   {
      AfficherMenu ();
      Choix = LireChoix ();
   }
   while (!ChoixEstValide (Choix));
   return Choix;
}

// Traite le choix reu en paramtre (valid) 
// l'aide du pointeur de IConfigTemps (valide)
void TraiterChoix (IConfigTemps *pConfigTemps, const char Choix)
{
   switch (Choix)
   {
      case CHOIX_TICSPARUNITE:
      {
         long NouvelleValeurTicsParUnite =
            LireEntierPositif ("Nombre de tics par unit de temps dsirs? ");
         pConfigTemps->SetUniteDeTemps (NouvelleValeurTicsParUnite);
      }
      break;
      case CHOIX_UNITESPARETAPE:
      {
         long NouvelleValeurUnitesParEtape=
            LireEntierPositif ("Nombre d'units de temps par tape dsires? ");
         pConfigTemps->SetUnitesParEtape (NouvelleValeurUnitesParEtape);
      }
      break;
   }
}
