#define _WIN32_DCOM

#include <windows.h>
#include "../Commun/UUID_Fabriques.h" // CLSID_FabriqueServeurTempsQuiPasse
#include "../Commun/CompterTemps_h.h" // ITempsQuiPasse
#include <olectl.h>    			// Pour les interfaces de point de connexion
#include <string>
#include <iostream>
#include <ChargeurCOM.h> // ChargeurCOM
#include <ConversionChaines.h> // StringToWString()
#include "ComposantDeRappel.h" // ComposantDeRappel

// Fixe les paramtres de connexion au serveur
class OutilDeConnexion
{
   static const std::wstring LOOPBACK_IP, LOOPBACK_TEXTE;
   static const std::wstring CTX_LOCAL, CTX_DISTANT;
   static const COSERVERINFO COSERVERINFO_Defaut;
   static const COAUTHINFO COAUTHINFO_Defaut;
   std::wstring m_IpServeur;
   long m_CtxServeur;
   COAUTHINFO m_InfoAuth;
   COSERVERINFO m_InfoServeur;
   bool ContexteLocal () const { return GetContexteServeur () == CLSCTX_LOCAL_SERVER; }
   bool ContexteDistant () const { return GetContexteServeur () == CLSCTX_REMOTE_SERVER; }
public:
   long GetContexteServeur () const { return m_CtxServeur; }
   const std::wstring & GetDescrContexteServeur () const
   {
      return ContexteLocal ()? CTX_LOCAL : CTX_DISTANT;
   }
   std::wstring GetIp () const { return m_IpServeur; }
   COSERVERINFO *GetInfoServeur () { return &m_InfoServeur; }
   OutilDeConnexion (const std::wstring &Ip)
      : m_IpServeur (Ip.length ()? Ip : LOOPBACK_IP),
        m_CtxServeur (Ip.length ()? CLSCTX_REMOTE_SERVER : CLSCTX_LOCAL_SERVER),
        m_InfoAuth (COAUTHINFO_Defaut),
        m_InfoServeur (COSERVERINFO_Defaut)
   {
      m_InfoAuth.dwAuthnLevel = RPC_C_AUTHN_LEVEL_NONE;
      m_InfoAuth.dwAuthnSvc = RPC_C_AUTHN_NONE;
      m_InfoAuth.dwAuthzSvc = RPC_C_AUTHZ_NONE;
      m_InfoAuth.dwCapabilities = EOAC_NONE;
      m_InfoAuth.dwImpersonationLevel = RPC_C_IMP_LEVEL_IMPERSONATE;
      m_InfoServeur.pAuthInfo = &m_InfoAuth;
      m_InfoServeur.pwszName = const_cast<wchar_t*>(m_IpServeur.c_str ());
   }
};
const std::wstring
   OutilDeConnexion::LOOPBACK_IP = L"127.0.0.1",
   OutilDeConnexion::LOOPBACK_TEXTE = L"localhost";
const std::wstring
   OutilDeConnexion::CTX_LOCAL = L"CLSCTX_LOCAL_SERVER",
   OutilDeConnexion::CTX_DISTANT = L"CLSCTX_REMOTE_SERVER";
const COSERVERINFO
   OutilDeConnexion::COSERVERINFO_Defaut = { 0 };
const COAUTHINFO
   OutilDeConnexion::COAUTHINFO_Defaut = { 0 };

int main (int argc, char *argv[])
{
   ChargeurCOM ccom;

   // --- CONNEXION
   OutilDeConnexion odc (argc > 1 ? StringToWString (argv[1]) : L"");
   std::wcout << L"Tentative de connexion au serveur..." << std::endl
              << L"--> Adresse:  " << odc.GetIp () << std::endl
              << L"--> Contexte: " << odc.GetDescrContexteServeur () << std::endl;

   // --- SCURIT
   // Scurit minimum
   ::CoInitializeSecurity(0, -1, 0, 0, RPC_C_AUTHN_LEVEL_NONE,
                          RPC_C_IMP_LEVEL_IDENTIFY, 0, EOAC_NONE, 0);

   // Cration du client de rappel. On veut un ITempsQuiPasse *, peu
   // importe le dtail de son implmentation. Avoir t plus gentil,
   // j'aurais mis ce qui suit dans une fonction...
   // La raison pour laquelle il y a un bloc ici est pour dlimiter
   // la porte de pRappel, qui ne sert que trs localement. Voyons
   // cela comme une espce de mini fabrique...
   ITempsQuiPasse *pTempsQuiPasse = 0;
   {
      ComposantDeRappel *pRappel = new ComposantDeRappel ();
      pRappel->AddRef ();
      if (FAILED (pRappel->QueryInterface
                    (IID_ITempsQuiPasse, 
                     reinterpret_cast<void **> (&pTempsQuiPasse))))
      {
         std::cerr << "ERREUR interne (QueryInterface() du composant de rappel)" << std::endl;
         return -1;
      }
      pRappel->Release ();
   }

   // --- CONNEXION ET INSTANCIATION
   // Cration du serveur. On veut un IConnectionPointContainer* ...
   MULTI_QI mqi [1] = { &__uuidof(IConnectionPointContainer), 0, S_OK };
   HRESULT hr = ::CoCreateInstanceEx (CLSID_FabriqueServeurTempsQuiPasse, 0,
                                      odc.GetContexteServeur (),
                                      odc.GetInfoServeur (), 1, mqi);
   if (FAILED (hr))
   {
      std::cerr << "Erreur de fabrication du serveur. Code: " << hr
                << std::endl;
      return -2;
   }
   IConnectionPointContainer *piCPC = static_cast<IConnectionPointContainer *>(mqi[0].pItf);

   // Obtention du point de connexion. Une fois le conteneur obtenu, on
   // lui demande un pointeur vers le point de connexion pour notre interface...
   IConnectionPoint *pCP = 0;
   if (SUCCEEDED (piCPC->FindConnectionPoint (IID_ITempsQuiPasse, &pCP)))
   {
      // Avis au serveur: rappelle-moi!
      DWORD dwCookie;
      if (SUCCEEDED (pCP->Advise (pTempsQuiPasse, &dwCookie)))
      {
         std::cout << "Pressez une touche, puis ENTER pour terminer..."
                   << std::endl;
         char c;
         std::cin >> c;
         // Avis au serveur: ne me rappelle plus!
         pCP->Unadvise (dwCookie);
      }
      pCP->Release ();
      piCPC->Release ();
   }
   pTempsQuiPasse->Release ();
}